using FMODUnity;
using HarmonyLib;
using Steamworks;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;
using UnityEngine;

class Dredger_Network : MonoBehaviour_Network
{
    public const float FINISHED_PROGRESS = 100f;
    public const Messages MSG_DROP = (Messages)(-50);
    public const Messages MSG_RAISE = (Messages)(-51);
    public const Messages MSG_ADD_PROGRESS = (Messages)(-52);
    public const Messages MSG_ADD_ITEM = (Messages)(-53);
    public const Messages MSG_TAKE_ITEM = (Messages)(-54);

    protected Item_Base currentItem = null;
    protected bool isDropped = false;
    protected bool isDropping = false;
    protected bool isRising = false;
    protected bool isPlaced = false;
    protected Dredger_Basket basket;
    protected Dredger_Electric elec;
    protected float progress = 0;
    protected DredgerType dredgerType;
    protected Animator anim;
    protected Rigidbody bucket;
    protected SpringJoint joint;
    protected StudioEventEmitter eventEmitterWeighLoop;
    protected Transform ropeConnectionPoint;
    protected Raft_Network network;

    public DredgerType DredgerType
    {
        get
        {
            return this.dredgerType;
        }
    }

    public bool HasItem
    {
        get
        {
            return this.currentItem != null;
        }
    }

    public bool IsDropped
    {
        get
        {
            return this.isDropped;
        }
    }

    public bool IsRising
    {
        get
        {
            return this.isRising;
        }
    }

    public float Progress
    {
        get
        {
            return this.progress;
        }
    }


    // Handle block being placed.
    public virtual void OnBlockPlaced()
    {
        NetworkIDManager.AddNetworkID(this);
        this.network = ComponentManager<Raft_Network>.Value;
        this.isPlaced = true;

        // Here is where all the fiddly bits go to determine what to do.
        Battery batt = this.GetComponentInChildren<Battery>();
        IRaycastable controls;
        var controlsComp = this.GetComponentsInChildren<RaycastInteractable>()
                               .First(x => x.gameObject.name.Contains("Controls"));
        if (batt)
        {
            this.elec = controlsComp.gameObject.AddComponent<Dredger_Electric>();
            this.elec.batt = batt;
            this.elec.network = this;
            controls = this.elec;
            this.dredgerType = DredgerType.ELECTRIC;
        }
        else
        {
            // This should run *only if manual*
            var handle = controlsComp.gameObject.AddComponent<Dredger_Handle>();
            handle.network = this;
            controls = handle;
            this.dredgerType = DredgerType.MANUAL;
        }
        var ri = controlsComp.GetComponent<RaycastInteractable>();
        ri?.AddRaycastables(new IRaycastable[] { controls });
        // Find the basket raycast.
        var basketComp = this.GetComponentsInChildren<RaycastInteractable>()
                             .First(x => x.gameObject.name.Contains("Dredge_Output"));

        // Add the basket component.
        this.basket = basketComp.gameObject.AddComponent<Dredger_Basket>();
        this.basket.network = this;
        var ri2 = basketComp.GetComponent<RaycastInteractable>();
        ri2?.AddRaycastables(new IRaycastable[] { this.basket });
        this.joint = this.GetComponentInChildren<SpringJoint>();
        this.bucket = this.GetComponentsInChildren<Rigidbody>()
                          .First(x => x.gameObject.name.Contains("Dredger_Bucket_Model"));
        this.eventEmitterWeighLoop = this.GetComponentInChildren<StudioEventEmitter>();
        this.ropeConnectionPoint = this.GetComponentsInChildren<Transform>()
                                       .First(x => x.gameObject.name.Contains("RopeConnectionPoint"));
        this.anim = this.GetComponentInChildren<Animator>();
    }

    public void Update()
    {
        if (!this.isPlaced)
        {
            return;
        }

        if (this.isDropping)
        {
            float num = Vector3.Distance(this.ropeConnectionPoint.position, this.bucket.position);
            float num2 = 2f;
            float num3 = Mathf.Clamp(num * 0.1f, 0f, num2);
            this.SetAnimSpeed(num3, num3 / num2);
            this.joint.minDistance += 10f * Time.deltaTime;
            if (this.joint.minDistance >= 50 | this.bucket.position.y < GameManager.OceanBottom)
            {
                this.SetAnimSpeed(0f, 0f);
                this.isDropping = false;
            }
        }
        else if (this.isRising)
        {
            if (this.joint.minDistance > 0f)
            {
                this.joint.minDistance -= 10f * Time.deltaTime;
                if (this.joint.minDistance <= 0f)
                {
                    this.isRising = false;
                    this.SetAnimSpeed(0f, 0f);
                    return;
                }
            }
        }
        else if (this.HasItem && !this.basket.ItemInBasket)
        {
            this.basket.ShowItem(this.currentItem);
        }

        // Send all of our messages on the update.
        if (Raft_Network.IsHost)
        {
            this.network.RPC(this.MessagesStack, Target.Other, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
        }
        else
        {
            this.network.SendP2P(this.network.HostID, this.MessagesStack, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
        }
        this.Clean();
    }

    public override bool Deserialize(Message_NetworkBehaviour msg, CSteamID remoteID)
    {
        switch (msg.Type)
        {
            case Messages.Battery_Insert:
            case Messages.Battery_Take:
            case Messages.Battery_Update:
            case Messages.Battery_OnOff:
                if (this.elec)
                {
                    return this.elec.batt.OnBatteryMessage(msg);
                }
                return false;
            case MSG_DROP:
                this.DropDredger(false);
                this.Retransmit(msg, remoteID);
                return false;
            case MSG_RAISE:
                this.RaiseDredger(false);
                this.Retransmit(msg, remoteID);
                return false;
            case MSG_ADD_PROGRESS:
                Message_HazmatSuit message = msg as Message_HazmatSuit;
                if (message == null)
                {
                    return false;
                }
                if (Raft_Network.IsHost)
                {
                    this.AddProgress(message.duration);
                }
                else
                {
                    this.progress += message.duration;
                }
                return false;
            case MSG_ADD_ITEM:
                // The host is the only one who should send this message. Reject
                // this if we are the host.
                Message_Item message2 = msg as Message_Item;
                if (message2 == null)
                {
                    return false;
                }
                if (!Raft_Network.IsHost)
                {
                    // Get the item.
                    this.currentItem = ItemManager.GetItemByIndex(message2.itemIndex);

                    if (!this.isDropped && ! this.isRising)
                    {
                        // Special situation that *shouldn't* happen. Handle it
                        // anyways.
                        this.basket.ShowItem(this.currentItem);
                    }
                }
                return false;
            case MSG_TAKE_ITEM:
                // Someone else picked up the item, so this is fine.
                this.PickupItem(null);
                this.Retransmit(msg, remoteID);
                return false;
            default:
                return base.Deserialize(msg, remoteID);
        }
    }

    public void SetAnimSpeed(float animSpeed, float soundSpeed)
    {
        this.anim?.SetFloat("Speed", animSpeed);
        this.eventEmitterWeighLoop?.SetParameter("velocity", soundSpeed);
        if (this.eventEmitterWeighLoop != null && soundSpeed == 0f && this.eventEmitterWeighLoop.IsPlaying())
        {
            this.eventEmitterWeighLoop.Stop();
            return;
        }
        if (this.eventEmitterWeighLoop != null && soundSpeed > 0f && !this.eventEmitterWeighLoop.IsPlaying())
        {
            this.eventEmitterWeighLoop.Play();
        }
    }

    // Handle the destruction of the block, either from unload or removing it.
    protected override void OnDestroy()
    {
        NetworkIDManager.RemoveNetworkID(this);
        base.OnDestroy();
        if (this.bucket != null)
        {
            Destroy(this.bucket.gameObject);
        }
    }

    protected void Retransmit(Message_NetworkBehaviour msg, CSteamID sourceID)
    {
        if (Raft_Network.IsHost)
        {
            foreach (KeyValuePair<CSteamID, Network_Player> kvPair in this.network.remoteUsers)
            {
                if (kvPair.Key.IsValid() && kvPair.Key != this.network.HostID && kvPair.Key != sourceID)
                {
                    this.network.SendP2P(kvPair.Key, msg, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
                }
            }
        }
    }

    // How we load the data from a save along with the harmony patch.
    public void RestoreData(RGD_Storage rgd)
    {
        if (rgd.isOpen)
        {
            this.DropDredger(false);
        }
        // Restore the current progress.
        this.progress = BitConverter.ToSingle(BitConverter.GetBytes(rgd.storageObjectIndex), 0);

        // Restore the current item and whether it is in the basket.

        this.currentItem = rgd.slots[0].slotIndex == 0 ? null : ItemManager.GetItemByIndex(rgd.slots[0].itemIndex);
        if (this.currentItem != null && rgd.slots[0].itemAmount == 1)
        {
            this.basket.ShowItem(this.currentItem);
        }

        if (this.elec)
        {
            if (rgd.slots[0].itemUses == 1) // legacy save loading
            {
                byte[] data = Convert.FromBase64String(rgd.slots[0].exclusiveString);
                int uses = BitConverter.ToInt32(data, 0);
                int index = BitConverter.ToInt32(data, 4);
                this.elec.batt.Insert(null, uses, index);
            }
            else if (rgd.slots[0].itemUses == 2)
                JsonUtility.FromJson<RGD_Battery>(rgd.slots[0].exclusiveString).RestoreBattery(elec.batt);
        }
    }

    public override RGD Serialize_Save()
    {
        RGD_Storage rgd = CreateObject<RGD_Storage>();
        rgd.CopyFieldsOf(new RGD_Block(RGDType.Block, this.GetComponent<Block>()));
        rgd.isOpen = this.isDropped;
        rgd.storageObjectIndex = BitConverter.ToUInt32(BitConverter.GetBytes(this.progress), 0);
        rgd.slots = new RGD_Slot[] { CreateObject<RGD_Slot>() };
        // Handle current item and basket.
        rgd.slots[0].slotIndex = (short)(this.currentItem == null ? 0 : 1);
        rgd.slots[0].itemAmount = (short)(this.basket.ItemInBasket ? 1 : 0);
        rgd.slots[0].itemIndex = (short)(this.currentItem == null ? 0 : this.currentItem.UniqueIndex);
        // Handle electric version's battery.
        if (this.elec != null)
        {
            if (this.elec.batt == null || this.elec.batt.GetBatteryInstance() == null)
                rgd.slots[0].itemUses = 0;
            else
            {
                rgd.slots[0].itemUses = 2;
                rgd.slots[0].exclusiveString = JsonUtility.ToJson(new RGD_Battery(elec.batt), false);
            }
        }

        return rgd;
    }

    public void DropDredger(bool sendMessage = true)
    {
        // Handle dropping it.
        if (this.isDropped)
        {
            Debug.LogError("Script tried to drop the dredger while it was already dropped.");
            var trace = new System.Diagnostics.StackTrace();
            var message = "";
            foreach (var frame in trace.GetFrames())
            {
                var method = frame.GetMethod();
                if (method.Name.Equals("LogStack"))
                    continue;
                message += string.Format(
                        "{0}::{1}",
                        method.ReflectedType != null ? method.ReflectedType.Name : string.Empty,
                        method.Name
                    ) + "\n";
            }
            Debug.Log(message);
            return;
        }
        if (sendMessage)
        {
            this.AddMessage(new Message_NetworkBehaviour(MSG_DROP, this));
        }
        this.isDropped = true;
        this.isDropping = true;
        this.bucket.transform.SetParent(null);
    }

    public void RaiseDredger(bool sendMessage = true)
    {
        // Handle raising it.
        if (sendMessage)
        {
            this.AddMessage(new Message_NetworkBehaviour(MSG_RAISE, this));
        }
        this.isDropped = false;
        this.isDropping = false;
        this.progress = 0;
        this.isRising = true;
        this.bucket.transform.SetParent(this.transform.parent);
        this.SetAnimSpeed(-0.5f, 1f);
    }

    public void AddProgress(float progress)
    {
        if (this.currentItem != null)
        {
            return;
        }

        if (Raft_Network.IsHost)
        {
            switch (this.dredgerType)
            {
                case DredgerType.MANUAL:
                    progress *= Dredgers.ManualProgressPerSecond;
                    break;
                case DredgerType.ELECTRIC:
                    progress *= Dredgers.ElectricProgressPerSecond;
                    break;
                default:
                // Shouldn't happen.
                    Debug.LogError("Error, unknown dredger type!");
                    return;
            }
            this.progress += progress;
            this.AddMessage(new Message_HazmatSuit(MSG_ADD_PROGRESS, this, false, progress));
            if (this.progress >= FINISHED_PROGRESS)
            {
                this.currentItem = this.basket.GetRandomItem();
                this.AddMessage(new Message_Item(MSG_ADD_ITEM, this, this.currentItem));
                this.RaiseDredger();
            }
        }
        else
        {
            this.AddMessage(new Message_HazmatSuit(MSG_ADD_PROGRESS, this, false, progress));
        }

    }

    public void PickupItem(Network_Player player)
    {
        if (this.currentItem != null)
        {
            if (player != null && player.IsLocalPlayer)
            {
                player.Inventory.AddItem(this.currentItem.UniqueName, 1);
                this.AddMessage(new Message_NetworkBehaviour(MSG_TAKE_ITEM, this));
            }
            this.basket.ClearItem();
        }

        // Clear the item at the end.
        this.currentItem = null;
    }

    public static T CreateObject<T>() => (T)FormatterServices.GetUninitializedObject(typeof(T));
}

class Dredger_Handle : MonoBehaviour, IRaycastable
{
    protected bool showingText = false;
    protected bool loadCircleShown = false;

    public Dredger_Network network;

    void IRaycastable.OnIsRayed()
    {
        if (this.network.IsDropped)
        {
            this.network.SetAnimSpeed(0.05f, 0.05f);
        }
        // If there is no item or it is at the bottom.
        CanvasHelper canvas = ComponentManager<CanvasHelper>.Value;
        if (!this.network.HasItem || this.network.IsDropped)
        {
            if (this.network.IsDropped)
            {

                this.showingText = true;
                this.loadCircleShown = true;
                canvas.displayTextManager.ShowText("Dredge");
                canvas.SetLoadCircle(true);
                canvas.SetLoadCircle(this.network.Progress / Dredger_Network.FINISHED_PROGRESS);
                this.network.AddProgress(Time.deltaTime);
            }
            else
            {
                if (this.loadCircleShown)
                {
                    canvas.SetLoadCircle(false);
                    this.loadCircleShown = false;
                }
                // Display message to drop it.
                canvas.displayTextManager.ShowText("Drop Dredger", MyInput.Keybinds["Interact"].MainKey);
                this.showingText = true;
                if (MyInput.GetButtonDown("Interact"))
                {
                    this.network.DropDredger();
                    this.showingText = false;
                }
            }
        }
        else
        {
            if (this.loadCircleShown)
            {
                canvas.SetLoadCircle(false);
                this.loadCircleShown = false;
            }
            if (this.network.IsRising)
            {
                canvas.displayTextManager.ShowText("Dredger is coming back up.");
            }
            else
            {
                // Some kind of message to say to pick up the item.
                canvas.displayTextManager.ShowText("Pick up the item in the basket!");
            }
        }
    }

    void IRaycastable.OnRayEnter() { }

    void IRaycastable.OnRayExit()
    {
        if (this.showingText)
        {
            ComponentManager<CanvasHelper>.Value.displayTextManager.HideDisplayTexts();
            this.showingText = false;
        }
        if (this.loadCircleShown)
        {
            ComponentManager<CanvasHelper>.Value.SetLoadCircle(false);
            this.loadCircleShown = false;
        }
        if (this.network.IsDropped)
        {
            this.network.SetAnimSpeed(0f, 0f);
        }
    }
}

class Dredger_Electric : MonoBehaviour, IRaycastable
{
    protected int batteryUsesPerMin = 4;
    protected bool loadCircleShown = false;
    protected float timer = 0;

    public Battery batt;
    public Dredger_Network network;

    void Start()
    {
        if (this.batt)
        {
            Traverse.Create(this.batt).Field("networkBehaviourID").SetValue(this.network);
        }
    }

    void Update()
    {
        if (Raft_Network.IsHost)
        {
            if (this.batt.CanGiveElectricity)
            {
                if (this.network.IsDropped)
                {
                    this.network.AddProgress(Time.deltaTime);
                    this.timer += Time.deltaTime;
                    float timePerBattery = 60f / this.batteryUsesPerMin;
                    if (this.timer >= timePerBattery)
                    {
                        // Handle multiple draws per second nicely.
                        int count = (int)(this.timer / timePerBattery);
                        this.timer -= count * timePerBattery;
                        this.batt.RemoveBatteryUsesNetworked(count);
                    }
                }
                else if (!this.network.HasItem)
                {
                    this.network.DropDredger();
                }
            }
        }
    }

    void IRaycastable.OnIsRayed()
    {
        CanvasHelper canvas = ComponentManager<CanvasHelper>.Value;
        if (this.network.IsDropped)
        {
            this.loadCircleShown = true;
            canvas.SetLoadCircle(true);
            canvas.SetLoadCircle(this.network.Progress / Dredger_Network.FINISHED_PROGRESS);
        }
        else
        {
            if (this.loadCircleShown)
            {
                canvas.SetLoadCircle(false);
                this.loadCircleShown = false;
            }
        }
    }

    void IRaycastable.OnRayEnter() { }

    void IRaycastable.OnRayExit()
    {
        if (this.loadCircleShown)
        {
            ComponentManager<CanvasHelper>.Value.SetLoadCircle(false);
            this.loadCircleShown = false;
        }
    }
}

class Dredger_Basket : MonoBehaviour, IRaycastable
{
    protected bool itemInBasket;
    protected bool showingText;
    protected ItemObjectEnabler itemVisual;

    public Dredger_Network network;

    public bool ItemInBasket
    {
        get
        {
            return this.itemInBasket;
        }
    }

    void Start()
    {
        this.showingText = false;
        this.itemInBasket = false;
        this.itemVisual = this.GetComponent<ItemObjectEnabler>();
    }

    // Clear the item from the basket.
    public void ClearItem()
    {
        this.itemInBasket = false;
        this.itemVisual?.DeActivateAllModels();

        if (this.showingText)
        {
            ComponentManager<CanvasHelper>.Value.displayTextManager.HideDisplayTexts();
            this.showingText = false;
        }
    }

    public Item_Base GetRandomItem()
    {
        return Dredgers.GetRandomDropper(this.network.DredgerType).GetRandomItem();
    }

    // Show the correct item.
    public void ShowItem(Item_Base item)
    {
        this.itemInBasket = true;
        // Turn everything off then turn on just the one we want. Handles the
        // component not properly existing.
        this.itemVisual?.DeActivateAllModels();
        this.itemVisual?.ActivateModel(item);
    }

    void IRaycastable.OnIsRayed()
    {
        if (this.itemInBasket)
        {
            if (!this.showingText)
            {
                // Show the text for picking up the item.
                ComponentManager<CanvasHelper>.Value.displayTextManager.ShowText(
                    "Pick up Item",
                    MyInput.Keybinds["Interact"].MainKey
                );
                this.showingText = true;
            }

            if (MyInput.GetButtonDown("Interact"))
            {
                this.network.PickupItem(ComponentManager<Network_Player>.Value);
            }
        }
    }

    void IRaycastable.OnRayEnter() { }

    void IRaycastable.OnRayExit()
    {
        if (this.showingText)
        {
            ComponentManager<CanvasHelper>.Value.displayTextManager.HideDisplayTexts();
            this.showingText = false;
        }
    }
}

public enum DredgerType
{
    MANUAL,
    ELECTRIC
}
