﻿using HarmonyLib;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;
using System.Reflection;
using UnityEngine;
using I2.Loc;

//Todo:
//rope visuals
//unload items better
//define most materials with c# to cut down mod size by 90% and solve the pale textures issue

public class Dredgers : Mod
{
    static float progress_per_second_manual;
    public static float ManualProgressPerSecond =>
        ExtraSettingsAPI_Loaded ? progress_per_second_manual : 5f;
    static float progress_per_second_electric;
    public static float ElectricProgressPerSecond =>
        ExtraSettingsAPI_Loaded ? progress_per_second_electric : 5f;
    static bool ExtraSettingsAPI_Loaded = false;
    static string dropperDifficulty;
    public static string DropperDifficulty =>
        ExtraSettingsAPI_Loaded ? DropperDifficulty : "Medium";

    void ExtraSettingsAPI_Load()
    {
        ExtraSettingsAPI_SetComboboxContent("dropperDifficulty", difficulties.ToArray());
        ExtraSettingsAPI_SettingsClose();
    }

    public virtual float ExtraSettingsAPI_GetSliderValue(string SettingName) => 0;

    public string ExtraSettingsAPI_GetComboboxSelectedItem(string SettingName) => "";

    void ExtraSettingsAPI_SettingsClose()
    {
        progress_per_second_manual = ExtraSettingsAPI_GetSliderValue("progress_per_second_manual");
        progress_per_second_electric = ExtraSettingsAPI_GetSliderValue(
            "progress_per_second_electric"
        );
        dropperDifficulty = ExtraSettingsAPI_GetComboboxSelectedItem("dropperDifficulty");
    }

    public static void ExtraSettingsAPI_ResetAllSettings() { }

    public void ExtraSettingsAPI_ButtonPress(string name)
    {
        if (name == "default")
        {
            ExtraSettingsAPI_ResetAllSettings();
        }
    }

    public void ExtraSettingsAPI_SetComboboxContent(string SettingName, string[] value) { }

    public static RandomDropper GetRandomDropper(DredgerType dt)
    {
        return dropperDict[(CurrentDropperType, dt)];
    }

    public static AssetBundle asset;
    public LanguageSourceData language;
    private Harmony harmony;
    protected static Dictionary<(string, DredgerType), RandomDropper> dropperDict =
        new Dictionary<(string, DredgerType), RandomDropper>();
    protected static List<string> difficulties = new List<string>();
    protected static String currentDropperType = "Medium";
    protected static String CurrentDropperType =>
        ExtraSettingsAPI_Loaded ? currentDropperType : "Medium";
    protected GameObject randomDroppersPrefab;

    public void Start()
    {
        try
        {
            language = new LanguageSourceData()
            {
                mDictionary = new Dictionary<string, TermData>
                {
                    ["CraftingSub/Dredgers"] = new TermData() { Languages = new[] { "Dredgers" } }
                },
                mLanguages = new List<LanguageData>
                {
                    new LanguageData() { Code = "en", Name = "English" }
                }
            };
            LocalizationManager.Sources.Add(language);

            HNotification notification = FindObjectOfType<HNotify>()
                .AddNotification(HNotify.NotificationType.spinning, "Loading Dredgers...");

            var bundleLoadRequest = AssetBundle.LoadFromMemoryAsync(
                GetEmbeddedFileBytes("dredgers.assets")
            );
            asset = bundleLoadRequest.assetBundle;

            var Dredger_Manual = asset.LoadAsset<Item_Base>("Placeable_Dredger_Manual");
            var Dredger_Electric = asset.LoadAsset<Item_Base>("Placeable_Dredger_Electric");

            RAPI.AddItemToBlockQuadType(Dredger_Manual, RBlockQuadType.quad_foundation);
            RAPI.AddItemToBlockQuadType(Dredger_Manual, RBlockQuadType.quad_foundation_empty);
            RAPI.RegisterItem(Dredger_Manual);
            RAPI.AddItemToBlockQuadType(Dredger_Electric, RBlockQuadType.quad_foundation);
            RAPI.AddItemToBlockQuadType(Dredger_Electric, RBlockQuadType.quad_foundation_empty);
            RAPI.RegisterItem(Dredger_Electric);

            LocalizationManager.LocalizeAll();

            foreach (Block b in Dredger_Manual.settings_buildable.GetBlockPrefabs())
            {
                b.networkedBehaviour = b.gameObject.AddComponent<Dredger_Network>();
                b.networkType = NetworkType.NetworkBehaviour;
            }
            foreach (Block b in Dredger_Electric.settings_buildable.GetBlockPrefabs())
            {
                b.networkedBehaviour = b.gameObject.AddComponent<Dredger_Network>();
                b.networkType = NetworkType.NetworkBehaviour;
            }

            Dredger_Manual.SetRecipe(
                new[]
                {
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(21) }, 10),
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(22) }, 10),
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(20) }, 10),
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(23) }, 15)
                },
                CraftingCategory.Other,
                1,
                false,
                "Dredgers",
                1
            );
            Dredger_Electric.SetRecipe(
                new[]
                {
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(21) }, 10),
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(22) }, 10),
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(70) }, 6),
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(307) }, 4)
                },
                CraftingCategory.Other,
                1,
                false,
                "Dredgers",
                2
            );

            this.harmony = new Harmony("com.plateglassarmour.Dredgers");
            this.harmony.PatchAll(Assembly.GetExecutingAssembly());

            notification.Close();
            Debug.Log("Dredgers has been successfully loaded!");

            this.randomDroppersPrefab = asset.LoadAsset<GameObject>("RandomDroppers");

            foreach (Transform child in this.randomDroppersPrefab.transform)
            {
                difficulties.Add(child.gameObject.name);
                foreach (
                    RandomDropper dropper in child.GetComponentsInChildren<RandomDropper>(true)
                )
                {
                    switch (dropper.gameObject.name)
                    {
                        case "DM":
                            dropperDict.Add((child.gameObject.name, DredgerType.MANUAL), dropper);
                            break;
                        case "DE":
                            dropperDict.Add((child.gameObject.name, DredgerType.ELECTRIC), dropper);
                            break;
                    }
                }
            }
        }
        catch (Exception e)
        {
            Debug.LogError(e);
        }
    }

    public void OnModUnload()
    {
        asset.Unload(true);
        this.harmony?.UnpatchAll("com.plateglassarmour.Dredgers");
        LocalizationManager.Sources.Remove(language);
        if (this.randomDroppersPrefab)
        {
            Destroy(this.randomDroppersPrefab);
        }
        Debug.Log("Dredgers has been unloaded!");
    }
}

static class ExtentionMethods
{
    public static void SetRecipe(
        this Item_Base item,
        CostMultiple[] cost,
        CraftingCategory category = CraftingCategory.Resources,
        int amountToCraft = 1,
        bool learnedFromBeginning = false,
        string subCategory = null,
        int subCatergoryOrder = 0
    )
    {
        Traverse recipe = Traverse.Create(item.settings_recipe);
        recipe.Field("craftingCategory").SetValue(category);
        recipe.Field("amountToCraft").SetValue(amountToCraft);
        recipe.Field("learnedFromBeginning").SetValue(learnedFromBeginning);
        recipe.Field("subCategory").SetValue(subCategory);
        recipe.Field("subCatergoryOrder").SetValue(subCatergoryOrder);
        item.settings_recipe.NewCost = cost;
    }
}

[HarmonyPatch(typeof(RGD_Block), "RestoreBlock")]
public class Patch_DredgerRestoreBlock
{
    public static void Postfix(ref RGD_Block __instance, Block block)
    {
        if (__instance.BlockIndex == 16952 || __instance.BlockIndex == 16953)
        {
            RGD_Storage rgd = __instance as RGD_Storage;
            Dredger_Network dn = block.networkedBehaviour as Dredger_Network;
            if (rgd != null && dn != null)
            {
                dn.RestoreData(rgd);
            }
        }
    }
}

[HarmonyPatch(typeof(Block), "GetBlockCreationData")]
public class Patch_BlockCreateSaveData
{
    public static bool Prefix(ref Block __instance, ref RGD_Block __result)
    {
        if (__instance.GetComponent<Dredger_Network>() != null)
        {
            __result = __instance.GetComponent<Dredger_Network>().Serialize_Save() as RGD_Block;
            return false;
        }
        return true;
    }
}

public static class ExtensionMethods
{
    // Thanks to aidan for this one.
    public static void CopyFieldsOf(this object value, object source)
    {
        var t1 = value.GetType();
        var t2 = source.GetType();
        while (!t1.IsAssignableFrom(t2))
            t1 = t1.BaseType;
        while (t1 != typeof(UnityEngine.Object) && t1 != typeof(object))
        {
            foreach (var f in t1.GetFields(~BindingFlags.Default))
                if (!f.IsStatic)
                    f.SetValue(value, f.GetValue(source));
            t1 = t1.BaseType;
        }
    }
}
